/**
 * @file  IntelHex.cpp
 * @brief Implementierung der Klasse CIntelHex.
 */
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "IntelHex.h"

#include <ctype.h>
#include <string.h>
#include <stdio.h>

int CIntelHex::m_errorState = 0;
char CIntelHex::m_errorString[];
char CIntelHex::m_errorText[];
char CIntelHex::m_errorDetail[];

CIntelHex::CIntelHex()
{
	// Member initialisieren
	ResetError();
	Reset();
}

CIntelHex::~CIntelHex()
{

}

void CIntelHex::ResetError()
{
	// Member initialisieren
	m_errorState = 0;
	memset(m_errorText, 0, sizeof(m_errorText));
	memset(m_errorDetail, 0, sizeof(m_errorDetail));
}

void CIntelHex::Reset()
{
	// Member initialisieren
}

// ########################################################
// HELPER

unsigned short CIntelHex::HexToUINT16(const char *data)
{
	int i;
	unsigned short val = 0;

	for (i=0; i < 4; i++) {

		val = (val << 4);

		if (data[i] >= '0' && data[i] <= '9') {

			val += data[i] - '0';

		} else if (data[i] >= 'a' && data[i] <='f') {

			val += data[i] - 'a' + 10;

		} else {

			val += data[i] - 'A' + 10;
		}
	}

	return val;
}

void CIntelHex::UINT16ToHex(unsigned short val, char *data)
{
	int i;

	for (i=3; i >= 0; i--) {

		data[i] = val & 0xF;

		if (data[i] > 9) {
			
			data[i] += 'A' - 10; 
		} else {
		
			data[i] += '0';
		}

		val >>= 4;
	}	
}

unsigned char CIntelHex::HexToUINT8(const char *data)
{
	int i;
	unsigned char val = 0;

	for (i=0; i < 2; i++) {

		val = (val << 4);

		if (data[i] >= '0' && data[i] <= '9') {

			val += data[i] - '0';

		} else if (data[i] >= 'a' && data[i] <='f') {

			val += data[i] - 'a' + 10;

		} else {

			val += data[i] - 'A' + 10;
		}
	}

	return val;
}

void CIntelHex::UINT8ToHex(unsigned char val, char *data)
{
	data[0] = (val >> 4) & 0xF;

	if (data[0] > 9) {
		
		data[0] += 'A' - 10; 
	} else {
		
		data[0] += '0';
	}

	data[1] = val & 0xF;

	if (data[1] > 9) {
		
		data[1] += 'A' - 10;
	} else {
		
		data[1] += '0';
	}
}

int CIntelHex::MakeCrc(char *data)
{
	int length = 0, lastByte, check = 0;

	if (data[length] != INTELHEX_START_SIGN) {

		return 0;
	}
	length++;

	// Laenge der Datenzeile ermitteln
	while(isxdigit((int)data[length])) {
		
		length++;

		if (!(length % 2)) {

			lastByte = HexToUINT8(&data[length - 1]);
			check += lastByte;
		}
	}

	if (length < 9) { // Mindestens 9 Zeichen: ":00000001"
		
		return 0;
	}

	// Pruefsumme eintragen
	UINT8ToHex((unsigned char)(~(((char*)&check)[0] - 1) & 0xFF), &data[length]);

	return 1;
}

int CIntelHex::CheckCrc(const char *data)
{
	int length = 0, lastByte = 0, check = 0;

	if (data[length] != INTELHEX_START_SIGN) {

		return 0;
	}
	length++;

	// Laenge der Datenzeile ermitteln
	while(isxdigit((int)data[length])) {
		
		length++;

		if (!(length % 2)) {

			lastByte = HexToUINT8(&data[length - 1]);
			check += lastByte;
		}
	}

	if (length < 11) { // Mindestens 11 Zeichen: ":00000001FF"
		
		return 0;
	}

	// Pruefsumme aus Gesamtsumme entfernen
	check -= lastByte;
	// Zweierkomplement
	check = ~(((char*)&check)[0] - 1) & 0xFF;
	
	return (check == lastByte);
}

int CIntelHex::MakeTypedDataLine(unsigned short address, unsigned char type, unsigned char *buf, unsigned char length, char *data)
{
	data[INTELHEX_START] = INTELHEX_START_SIGN;    // start of record
	UINT8ToHex(length, &data[INTELHEX_COUNT]);     // byte count
	UINT16ToHex(address, &data[INTELHEX_ADDRESS]); // load address
	UINT8ToHex(type, &data[INTELHEX_TYPE]);        // record type

	// insert data
	unsigned char i;
	for (i=0; i < length; i++) {

		UINT8ToHex(buf[i], &data[INTELHEX_DATA + (i * 2)]);
	}

	// Endezeichen fuer CRC - Laengenermittlung
	data[INTELHEX_DATA + (i * 2)] = 0;

	// checksum suffix
	if (MakeCrc(data)) {

		i++;

		// Nullterminie
		data[INTELHEX_DATA + (i * 2)] = 0;
		return 1;
	}

	return 0;
}

int CIntelHex::MakeDataLine(unsigned short address, unsigned char *buf, unsigned char length, char *data)
{
	return MakeTypedDataLine(address, INTELHEX_TYPE_DAT, buf, length, data);
}

int CIntelHex::MakeEndLine(char *data)
{
	data[INTELHEX_START] = INTELHEX_START_SIGN;          // start of record
	UINT8ToHex(0, &data[INTELHEX_COUNT]);                // byte count
	UINT16ToHex(0, &data[INTELHEX_ADDRESS]);             // load address
	UINT8ToHex(INTELHEX_TYPE_EOF, &data[INTELHEX_TYPE]); // record type

	// Endezeichen fuer CRC - Laengenermittlung
	data[INTELHEX_DATA] = 0;

	// checksum suffix
	if (MakeCrc(data)) {

		// Nullterminie
		data[INTELHEX_DATA + 2] = 0;
		return 1;
	}

	return 0;
}

// ########################################################
// Sucht den Anfang und fuehrt grundlegende Pruefungen durch.
char *CIntelHex::PrepareDataLine(char *data, int length)
{
	int pos;
	// Datensatzanfang suchen
	for (pos=0; data[pos] != '\0' && pos < length; pos++) {

		// Startzeichen?
		if (data[pos] == INTELHEX_START_SIGN) {

			// Liegt der gesamte Datensatz im Puffer vor?
			if ((INTELHEX_LENGTH_FRAME + GetCount(&data[pos])*2) > (length - pos)) {

				m_errorState = INTELHEX_ES_RECORD_OVER_BUFFER;
				sprintf(m_errorDetail, "%s(%i): %s", __FILE__, __LINE__, INTELHEX_ES_RECORD_OVER_BUFFER_TEXT);
				return NULL;
			}

			// Checksumme pruefen
			if (!CheckCrc(&data[pos])) {

				m_errorState = INTELHEX_ES_CRC;
				sprintf(m_errorDetail, "%s(%i): %s", __FILE__, __LINE__, INTELHEX_ES_CRC_TEXT);
				return NULL;
			}

			return &data[pos];

		} else if (data[pos] != '\n' && data[pos] != '\t' && data[pos] != ' ') {

			// Zwischen den datenrecords duerfen nur Whitespaces liegen
			m_errorState = INTELHEX_ES_FORMAT_RECORD;
			sprintf(m_errorDetail, "%s(%i): %s", __FILE__, __LINE__, INTELHEX_ES_FORMAT_RECORD_TEXT);
			return NULL;
		}
	}
	
	m_errorState = INTELHEX_ES_EMPTY;
	sprintf(m_errorDetail, "%s(%i): %s", __FILE__, __LINE__, INTELHEX_ES_EMPTY_TEXT);
	return NULL;
}

// ########################################################
// Laden und Pruefen der Datei.
int CIntelHex::CheckContent(const char *fileName)
{
	size_t pos = 0;
	unsigned int numRecord = 0;
	char buf[530], *cur = NULL, *curTmp;

	FILE *fp = fopen(fileName, "r");

	if (fp == NULL) {
	
		m_errorState = INTELHEX_ES_FILE_OPEN;
		sprintf(m_errorDetail, "%s(%i): %s", __FILE__, __LINE__, INTELHEX_ES_FILE_OPEN_TEXT);
		return 0;
	}

	// Leerer Datenpuffer als Vorgabe
	m_errorState = INTELHEX_ES_EMPTY;
	sprintf(m_errorDetail, "%s(%i): %s", __FILE__, __LINE__, INTELHEX_ES_EMPTY_TEXT);
	
	do {
		
		if (m_errorState == INTELHEX_ES_RECORD_OVER_BUFFER) {
			
			pos = sizeof(buf) - (cur - buf);
			// Datensatzrest an Pufferanfang kopieren
			memcpy(buf, cur, pos);
			// Datenpuffer leeren und fuellen
			memset(&buf[pos], 0, sizeof(buf) - pos);
			fread(&buf[pos], sizeof(buf) - pos, 1, fp);
			// Datenzeiger setzen
			cur = buf;
			
		} else if (m_errorState == INTELHEX_ES_EMPTY) {
			
			// Datenpuffer fuellen
			memset(buf, 0, sizeof(buf));
			fread(buf, sizeof(buf), 1, fp);
			// Datenzeiger setzen
			cur = buf;
		} else {
			
			// Es ist ein Verarbeitungsfehler eingetreten.
			break;
		}
		
		if(ferror(fp)) {
			
			// Lag ein Dateifehler vor?
			m_errorState = INTELHEX_ES_FILE_ERROR;
			sprintf(m_errorDetail, "%s(%i): %s", __FILE__, __LINE__, INTELHEX_ES_FILE_ERROR_TEXT);
			break;
		}
		
		// Alle gueltigen Datensaetze im Puffer pruefen
		while((curTmp = PrepareDataLine(cur, int(sizeof(buf) - (cur - buf)))) != NULL) {
			
			// Records zaehlen
			numRecord++;
			// Recordstart uebernehmen
			cur = curTmp;
			// Hinter aktuellen Datenrecord setzen
			cur += INTELHEX_LENGTH_FRAME + GetCount(cur)*2;
		}
	} while(!feof(fp));

	// Wurde die Datei komplett gelesen und der Puffer ist leer, dann 
	// ist die Datei gueltig, wenn mindestens ein Datensatz vorhanden war.
	if (feof(fp) && m_errorState == INTELHEX_ES_EMPTY && numRecord) {
	
		m_errorState = INTELHEX_ES_NO;
		memset(m_errorDetail, 0, sizeof(m_errorDetail));
	} else if (m_errorState == INTELHEX_ES_EMPTY && numRecord == 0) {

		// Es liegt kein Datensatz vor, die Datei ist somit leer.
		m_errorState = INTELHEX_ES_FILE_EMPTY;
		sprintf(m_errorDetail, "%s(%i): %s", __FILE__, __LINE__, INTELHEX_ES_FILE_EMPTY_TEXT);
	}

	// Datei schliessen
	fclose(fp);

	return (m_errorState == INTELHEX_ES_NO);
}

// ########################################################
// Pruefen der Datei mit Inhalt.
int CIntelHex::CheckFile(const char *fileName)
{
	unsigned int pos = 0, numRecord = 0;
	char buf[530];
	unsigned char fail = 0;

	FILE *fp = fopen(fileName, "r");

	if (fp == NULL) {
	
		return 0;
	}

	memset(buf, 0, sizeof(buf));
	while(!feof(fp)) {
		
		fread(&buf[pos], 1, 1, fp);

		if (buf[pos] == INTELHEX_START_SIGN && numRecord) {
		
			if (!CheckCrc(buf)) {

				fail = 1;
				break;
			}

			memset(buf, 0, sizeof(buf));

			pos = 0;
			numRecord++; // Naechster Datensatz in Vorbereitung
			buf[0] = INTELHEX_START_SIGN;
		} 

		if (pos >= sizeof(buf)) {

			fail = 1;
			break;
		}
		// Erster Datensatz angefangen
		if (!numRecord) {
		
			numRecord++;
		}

		pos++;
	}

	fclose(fp);

	if (!fail) {

		if (!CheckCrc(buf)) {

			fail = 1;
		} else {

			// Abschlusszeile pruefen (End-of-file record) 
			if (GetType(buf) != INTELHEX_TYPE_EOF) {

				fail = 1;
			}
		}
	}

	return !fail;
}

unsigned char CIntelHex::GetCount(const char *data)
{
	return HexToUINT8(&data[INTELHEX_COUNT]);
}

unsigned short CIntelHex::GetAddress(const char *data)
{
	return HexToUINT16(&data[INTELHEX_ADDRESS]);
}

unsigned char CIntelHex::GetType(const char *data)
{
	return HexToUINT8(&data[INTELHEX_TYPE]);
}

unsigned char CIntelHex::GetData(const char *data, unsigned char *dest)
{// vom Datensatztyp unabhaengig
	unsigned short num, length = GetCount(data);
	if (length > 0xFF) {
	
		return 0;
	}

	for (num=0; num < length; num++) {
	
		dest[num] = HexToUINT8(&data[INTELHEX_DATA + 2 * num]);
	}

	return (unsigned char)num;
}

unsigned char CIntelHex::GetData8(const char *data, unsigned char num)
{
	if (num > GetCount(data)) {
	
		return 0;
	}

	return HexToUINT8(&data[INTELHEX_DATA + 2 * num]);
}

unsigned short CIntelHex::GetData16(const char *data, unsigned char num)
{
	if ((num + 1) > GetCount(data)) {
	
		return 0;
	}

	return HexToUINT16(&data[INTELHEX_DATA + 2 * num]);
}

char *CIntelHex::GetErrorString(int errorState)
{
	// Ist der Parameter -1 wird die Fehlervariable der Klasse verwendet.
	if (errorState == -1) {

		errorState = m_errorState;
	}

	switch(errorState) {
	case INTELHEX_ES_NO:
		sprintf(m_errorString, "(%i): %s", errorState, INTELHEX_ES_NO_TEXT);
		break;
	case INTELHEX_ES_FILE_OPEN:
		sprintf(m_errorString, "(%i): %s", errorState, INTELHEX_ES_FILE_OPEN_TEXT);
		break;
	case INTELHEX_ES_FILE_FORMAT:
		sprintf(m_errorString, "(%i): %s", errorState, INTELHEX_ES_FILE_FORMAT_TEXT);
		break;
	case INTELHEX_ES_EMPTY:
		sprintf(m_errorString, "(%i): %s", errorState, INTELHEX_ES_EMPTY_TEXT);
		break;
	case INTELHEX_ES_CRC:
		sprintf(m_errorString, "(%i): %s", errorState, INTELHEX_ES_CRC_TEXT);
		break;
	case INTELHEX_ES_FORMAT_RECORD:
		sprintf(m_errorString, "(%i): %s", errorState, INTELHEX_ES_FORMAT_RECORD_TEXT);
		break;
	case INTELHEX_ES_RECORD_OVER_BUFFER:
		sprintf(m_errorString, "(%i): %s", errorState, INTELHEX_ES_RECORD_OVER_BUFFER_TEXT);
		break;
	case INTELHEX_ES_FILE_ERROR:
		sprintf(m_errorString, "(%i): %s", errorState, INTELHEX_ES_FILE_ERROR_TEXT);
		break;
	case INTELHEX_ES_FILE_EMPTY:
		sprintf(m_errorString, "(%i): %s", errorState, INTELHEX_ES_FILE_EMPTY_TEXT);
		break;
	default:
		sprintf(m_errorString, "(%i): %s", errorState, "Unbekannter Fehler.");
	}

	return m_errorString;
}
